/******************************************************************************
 * MODULE       : rohm_bh1780_i2c.c
 * FUNCTION     : Driver source for BH1780, Ambient Light Sensor(ALS) IC
 * AUTHOR       : Masafumi Seike <masafumi.seike@dsn.rohm.co.jp >
 * PROGRAMMED   : Sensor Application Development Group, ROHM CO.,LTD.
 * MODIFICATION : Modified by ROHM, JUN/08/2012
 * REMARKS      :
 * COPYRIGHT    : Copyright (C) 2012 ROHM CO.,LTD.
 *              : This program is free software; you can redistribute it and/or
 *              : modify it under the terms of the GNU General Public License
 *              : as published by the Free Software Foundation; either version 2
 *              : of the License, or (at your option) any later version.
 *              : 
 *              : This program is distributed in the hope that it will be useful,
 *              : but WITHOUT ANY WARRANTY; without even the implied warranty of
 *              : MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *              : GNU General Public License for more details.
 *              : 
 *              : You should have received a copy of the GNU General Public License
 *              : along with this program; if not, write to the Free Software
 *              : Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 *****************************************************************************/

#include <linux/module.h>
#include <linux/delay.h>
#include <linux/hrtimer.h>
#include <linux/i2c.h>
#include <linux/interrupt.h>
#include <linux/io.h>
#include <linux/gpio.h>
#include <linux/miscdevice.h>
#include <asm/irq.h>
#include <asm/uaccess.h>
#include <linux/platform_device.h>
#include <linux/input.h>
#include <linux/workqueue.h>
#include <linux/proc_fs.h>

#include "rohm_bh1780_i2c.h"
#include "rohm_bh1780_i2c_config.h"
#include "rohm_bh1780_i2c_if.h"

/******************************* define *******************************/
/* structure of peculiarity to use by system */
typedef struct {
    struct i2c_client   *client;     /* structure pointer for i2c bus */
    struct hrtimer      timer;       /* structure for timer handler */
    struct work_struct  work;        /* structure for work queue */
    struct input_dev    *input_dev;  /* structure pointer for input device */
    struct delayed_work input_work;  /* structure for work queue */
    int                 delay_time;  /* delay time to set from application */
    int                 fopen_cnt;   /* open count */
} ALS_DATA;

/* logical functions */
static void                 als_work_func(struct work_struct *work);
static enum hrtimer_restart als_timer_func(struct hrtimer *timer);
static int                  als_iodev_open(struct inode *inode, struct file *file);
static int                  als_iodev_release(struct inode *inode, struct file *file);
static int                  als_iodev_ioctl(struct inode *inode, struct file *file, unsigned int cmd, unsigned long arg);
static int                  als_ioctl_power_on_off(ALS_DATA *als, unsigned char power_data);
static int                  als_probe(struct i2c_client *client, const struct i2c_device_id *id);
static int                  als_remove(struct i2c_client *client);
static int __devinit        als_init(void);
static void __exit          als_exit(void);
/* access functions */
static int als_driver_init(void);
static int als_driver_shutdown(struct i2c_client *client);
static int als_driver_reset(struct i2c_client *client);
static int als_driver_power_on_off(unsigned char data, struct i2c_client *client);
static int als_driver_read_power_state(unsigned char *pwr_st, struct i2c_client *client);
static int als_driver_read_data(unsigned short *data, struct i2c_client *client);
static int als_driver_general_read(GENREAD_ARG data, struct i2c_client *client);

/**************************** variable declaration ****************************/
static const char              bh1780_driver_ver[] = BH1780_DRIVER_VER;
static struct workqueue_struct *rohm_workqueue;
static ALS_DATA                *als_ginfo;

/**************************** structure declaration ****************************/
/* I2C device IDs supported by this driver */
static const struct i2c_device_id als_id[] = {
    { BH1780_I2C_NAME, 0 }, /* rohm bh1780 driver */
    { }
};

/* represent an I2C device driver */
static struct i2c_driver bh1780_driver = {
    .driver = {                   /* device driver model driver */
        .name = BH1780_I2C_NAME,
    },
    .probe    = als_probe,        /* callback for device binding */
    .remove   = als_remove,       /* callback for device unbinding */
    .id_table = als_id,           /* list of I2C devices supported by this driver */
};

/* structure of file operation */
static struct file_operations bh1780_fops = {
    .owner   = THIS_MODULE,       /* module of local file system */
    .open    = als_iodev_open,    /* called function when application call open */
    .release = als_iodev_release, /* called function when application call close */
    .ioctl   = als_iodev_ioctl,   /* called function when application call ioctl */
};

/* structure of misc deveice */
static struct miscdevice bh1780_device = {
    .minor = MISC_DYNAMIC_MINOR,  /* the minor number being registered */
    .name  = "bh1780_iodev",      /* registered name to kernel */
    .fops  = &bh1780_fops,        /* registered file operation to kernel */
};

/************************************************************
 *                      logic function                      *
 ***********************************************************/
/******************************************************************************
 * NAME       : als_work_func
 * FUNCTION   : periodically reads the data from sensor(thread of work)
 * REMARKS    :
 *****************************************************************************/
static void als_work_func(struct work_struct *work)
{
    int            result;
    unsigned short als_data;
    ALS_DATA       *als;
    long           get_timer;
    long           wait_sec;
    unsigned long  wait_nsec;
    unsigned char  pwr_st;

    als_data = 0;
    als      = container_of(work, ALS_DATA, work);

    /* read the state of sensor */
    result = als_driver_read_power_state(&pwr_st, als->client);
    if (result < 0) {
        printk(KERN_ERR "power_state error \n");
        return;
    }
    /* check the state of sensor */
    if (pwr_st == 0) {
        return;
    }
    result = als_driver_read_data(&als_data, als->client);
    if (result < 0) {
        printk(KERN_ERR "ERROR! read data\n");
        return;
    }
    input_report_abs(als->input_dev, ABS_X, (int)als_data);
    input_sync(als->input_dev);

    get_timer = als->delay_time;
    /* the setting value from application */
    /* 125ms(8Hz) at least */
    wait_sec  = (get_timer / SM_TIME_UNIT);
    wait_nsec = ((get_timer - (wait_sec * SM_TIME_UNIT)) * MN_TIME_UNIT);
    result = hrtimer_start(&als->timer, ktime_set(wait_sec, wait_nsec), HRTIMER_MODE_REL);
    if (result != 0) {
        printk(KERN_ERR "can't start timer\n");
        return;
    }
}

/******************************************************************************
 * NAME       : als_timer_func
 * FUNCTION   : call work function (thread of timer)
 * REMARKS    :
 *****************************************************************************/
static enum hrtimer_restart als_timer_func(struct hrtimer *timer)
{
    ALS_DATA *als;
    int      result;

    als    = container_of(timer, ALS_DATA, timer);
    result = queue_work(rohm_workqueue, &als->work);
    if (result == 0) {
        printk(KERN_ERR "can't register que.\n");
        printk(KERN_ERR "result = 0x%x\n", result);
    }

    return (HRTIMER_NORESTART);
}

/******************************************************************************
 * NAME       : als_iodev_open
 * FUNCTION   : initialize device and open process
 * REMARKS    :
 *****************************************************************************/
static int als_iodev_open(struct inode *inode, struct file *file)
{
#define FOPEN_MAX_TIME (1)

    unsigned char pwr_st;
    int           result;
    ALS_DATA      *als;

    als = als_ginfo;
    if (als->fopen_cnt >= FOPEN_MAX_TIME) {
        printk(KERN_ERR "already call open function! open time = %d \n", als->fopen_cnt);
        return (-ENFILE);
    }
    als->fopen_cnt++;
    /* read power state */
    result = als_driver_read_power_state(&pwr_st, als->client);
    if (result < 0) {
        printk(KERN_ERR "power_state error \n");
        return (result);
    }
    /* check the stop system is */
    if (pwr_st == 0) {
        /* set initialize state */
        result = als_driver_init();

        if (result < 0) {
            printk(KERN_ERR "als_probe don't initialize to sensor\n");
            printk(KERN_ERR "error value = 0x%x\n", result);
            return (result);
        }
        file->private_data = als;
    } else {
        printk(KERN_ERR "already open device !!\n");
    }

    return (result);

#undef FOPEN_MAX_TIME
}

/******************************************************************************
 * NAME       : als_iodev_release
 * FUNCTION   : stop device and close process
 * REMARKS    :
 *****************************************************************************/
static int als_iodev_release(struct inode *inode, struct file *file)
{
    int      result;
    ALS_DATA *als;

    /* copy client data */
    als = file->private_data;

    /* close driver */
    als->fopen_cnt--;
    result = als_driver_shutdown(als->client);
    if (result != 0) {
        printk(KERN_ERR "can't close device\n");
    }

    return (result);
}

/******************************************************************************
 * NAME       : als_iodev_ioctl
 * FUNCTION   : set each parameter to device and read version and set timer
 * REMARKS    :
 *****************************************************************************/
#define GEN_READ_MAX (5)
unsigned char gen_data[GEN_READ_MAX];
static int als_iodev_ioctl(struct inode *inode, struct file *file, unsigned int cmd, unsigned long arg)
{
    void __user   *argp;
    int           result;
    unsigned long time;
    ALS_DATA      *als;
    unsigned char pwr_data;
    GENREAD_ARG   gen_read;

    argp = (void __user *)arg;
    als  = file->private_data;
    switch (cmd) {
    case IOCTL_APP_SET_TIMER:
        if (copy_from_user(&time, argp, sizeof(time))) {
            printk(KERN_ERR "parameter NG !!\n");
            return (-EFAULT);
        }
        if (time >= (ALS_SET_MIN_DELAY_TIME)) {
            als_ginfo->delay_time = time;
        } else {
            printk(KERN_ERR "parameter is too small. \n");
            return (-EFAULT);
        }
        break;

    case IOCTL_APP_SET_PWRSET_ALS:
        if (copy_from_user(&pwr_data, argp, sizeof(pwr_data))) {
            printk(KERN_ERR "parameter NG !!\n");
            return (-EFAULT);
        }
        /* input power data */
        result = als_ioctl_power_on_off(als, pwr_data);
        if (result < 0) {
            printk(KERN_ERR "error value = 0x%x \n", result);
            return (result);
        }
        break;

    case IOCTL_APP_READ_DRIVER_VER:
        if (copy_to_user(argp, &bh1780_driver_ver, sizeof(bh1780_driver_ver))) {
            printk(KERN_ERR "parameter NG !!\n");
            return (-EFAULT);
        }
        break;

    case IOCTL_APP_SET_GENERAL:
        if (copy_from_user(&gen_read, argp, sizeof(gen_read))) {
            printk(KERN_ERR "parameter NG !!\n");
            return (-EFAULT);
        }
        if (gen_read.size > GEN_READ_MAX) {
            printk(KERN_ERR "parameter NG!! size is the bigger than %d bytes.\n", GEN_READ_MAX);
            return (-EFAULT);
        }
        memset(&gen_data, 0, sizeof(gen_data));
        gen_read.addr = gen_data;
        result = als_driver_general_read(gen_read, als->client);
        if (result < 0) {
            printk(KERN_ERR "error value = 0x%x \n", result);
            return (result);
        }
        break;

    case IOCTL_APP_READ_GENERAL:
        if (copy_to_user(argp, &gen_data, sizeof(gen_data))) {
            printk(KERN_ERR "parameter NG !!\n");
            return (-EFAULT);
        }
        break;

    default:
        printk(KERN_ERR "non cmd parameter!\n");
        return (-EFAULT);
        break;
    }

    return (0);
}
#undef GEN_READ_MAX

/******************************************************************************
 * NAME       : als_ioctl_power_on_off
 * FUNCTION   : execute power on/off of device
 * REMARKS    :
 *****************************************************************************/
static int als_ioctl_power_on_off(ALS_DATA *als, unsigned char power_data)
{
    unsigned char pwr_st;
    int           result;

    result = als_driver_read_power_state(&pwr_st, als->client);
    if (result < 0) {
        printk(KERN_ERR "power_state error \n");
        return (result);
    }
    /* set power state */
    result = als_driver_power_on_off(power_data, als->client);
    if (result < 0) {
        printk(KERN_ERR "power on failed\n");
        printk(KERN_ERR "error value = 0x%x\n", result);
        return (result);
    }
    if (pwr_st == 0) {
        /* start timer of 1 second */
        result = hrtimer_start(&als_ginfo->timer, ktime_set(1, 0), HRTIMER_MODE_REL);
        if (result != 0) {
            printk(KERN_ERR "can't start timer\n");
            return (result);
        }
    }

    return (result);
}

/******************************************************************************
 * NAME       : als_probe
 * FUNCTION   : initialize system
 * REMARKS    :
 *****************************************************************************/
static int als_probe(struct i2c_client *client, const struct i2c_device_id *id)
{
#define ROHM_ALS_MAX (65535)

    ALS_DATA *als;
    int      result;

    printk(KERN_INFO "called als_probe for BH1780!!\n");
    result = i2c_check_functionality(client->adapter, I2C_FUNC_I2C);
    if (!result) {
        printk(KERN_ERR "need I2C_FUNC_I2C\n");
        result = -ENODEV;
        goto err_check_functionality_failed;
    }
    als = kzalloc(sizeof(*als), GFP_KERNEL);
    if (als == NULL) {
        result = -ENOMEM;
        goto err_alloc_data_failed;
    }
    INIT_WORK(&als->work, als_work_func);
    als->client = client;
    i2c_set_clientdata(client, als);

    /* set input device */
    INIT_DELAYED_WORK(&als->input_work, als_work_func);
    als->input_dev = input_allocate_device();
    if (!als->input_dev) {
        result = -ENOMEM;
        dev_err(&als->client->dev, "input device allocate failed\n");
        goto err_power_failed;
    }
    input_set_drvdata(als->input_dev, als);
    /* set event bit */
    set_bit(EV_ABS, als->input_dev->evbit);
    set_bit(EV_SYN, als->input_dev->evbit);

    /* initialization of abs event */
    input_set_abs_params(als->input_dev, ABS_X, 0, ROHM_ALS_MAX, 0, 0);

    /* set event name */
    als->input_dev->name = "bh1780_inputdev";

    /* register the device */
    result = input_register_device(als->input_dev);
    if (result) {
        dev_err(&als->client->dev,
                "unable to register input polled device %s: %d\n",
                als->input_dev->name, result);
        goto err_inputdev;
    }
    /* reset for sensor */
    result = als_driver_reset(als->client);
    if (result != 0) {
        printk(KERN_ERR "don't reset result = 0x%x\n", result);
        goto err_inputdev;
    }
    /* timer process */
    hrtimer_init(&als->timer, CLOCK_MONOTONIC, HRTIMER_MODE_REL);
    als->timer.function = als_timer_func;
    result = misc_register(&bh1780_device);
    if (result != 0) {
        printk(KERN_ERR "als_probe register failed\n");
        goto err_power_failed;
    }

    /* initialize static variable */
    als->delay_time = 0;
    als->fopen_cnt  = 0;
    als_ginfo       = als;

    return (result);
err_inputdev:
    input_free_device(als->input_dev);
err_power_failed:
    kfree(als);
err_alloc_data_failed:
err_check_functionality_failed:

    return (result);

#undef ROHM_ALS_MAX
}

/******************************************************************************
 * NAME       : als_remove
 * FUNCTION   : close system
 * REMARKS    :
 *****************************************************************************/
static int als_remove(struct i2c_client *client)
{
    ALS_DATA *als;
    int      result;

    als    = i2c_get_clientdata(client);
    result = hrtimer_cancel(&als->timer);
    if (result != 0) {
        printk(KERN_ERR "can't finish timer.\n");
        return (result);
    }
    kfree(als);
    result = misc_deregister(&bh1780_device);
    if (result != 0) {
        printk(KERN_ERR "misc driver don't register in list.\n");
    }

    return (result);
}

/******************************************************************************
 * NAME       : als_init
 * FUNCTION   : register driver to kernel
 * REMARKS    :
 *****************************************************************************/
static int __devinit als_init(void)
{
    rohm_workqueue = create_singlethread_workqueue("rohm_workqueue");
    if (!rohm_workqueue) {
        return (-ENOMEM);
    }

    return (i2c_add_driver(&bh1780_driver));
}

/******************************************************************************
 * NAME       : als_exit
 * FUNCTION   : remove driver from kernel
 * REMARKS    :
 *****************************************************************************/
static void __exit als_exit(void)
{
    i2c_del_driver(&bh1780_driver);
    if (rohm_workqueue) {
        destroy_workqueue(rohm_workqueue);
    }

    return;
}

/************************************************************
 *                     access function                      *
 ***********************************************************/
/******************************************************************************
 * NAME       : als_driver_init
 * FUNCTION   : initialize BH1780 (non process)
 * REMARKS    :
 *****************************************************************************/
static int als_driver_init(void)
{
    return (0);

}

/******************************************************************************
 * NAME       : als_driver_shutdown
 * FUNCTION   : shutdown BH1780
 * REMARKS    :
 *****************************************************************************/
static int als_driver_shutdown(struct i2c_client *client)
{
    int result;

    /* set soft ware reset */
    result = als_driver_reset(client);

    return (result);
}

/******************************************************************************
 * NAME       : als_driver_reset
 * FUNCTION   : reset BH1780 register
 * REMARKS    :
 *****************************************************************************/
static int als_driver_reset(struct i2c_client *client)
{
    int result;

    /* set soft ware reset */
    result = i2c_smbus_write_byte_data(client, REG_ALSCONTROL, CTL_STANDBY);

    return (result);
}

/******************************************************************************
 * NAME       : als_driver_power_on_off
 * FUNCTION   : power on and off BH1780
 * REMARKS    :
 *****************************************************************************/
static int als_driver_power_on_off(unsigned char data, struct i2c_client *client)
{
    unsigned char als_ctl;
    int           result;

    /* check the parameter of als power */
    if (data > ALS_ENABLE) {
        return (-EINVAL);
    }
    /* set ALS control register */
    if (data == ALS_ENABLE) {
        als_ctl = CTL_STANDALONE;
    } else {
        als_ctl = CTL_STANDBY;
    }
    /* write register to BH1780 via i2c */
    result = i2c_smbus_write_byte_data(client, REG_ALSCONTROL, als_ctl);

    return (result);
}

/******************************************************************************
 * NAME       : als_driver_read_power_state
 * FUNCTION   : read the value of ALS status in BH1780
 * REMARKS    :
 *****************************************************************************/
static int als_driver_read_power_state(unsigned char *pwr_st, struct i2c_client *client)
{
    int result;

    /* read register to BH1780 via i2c */
    result = i2c_smbus_read_byte_data(client, REG_ALSCONTROL);
    if (result < 0) {
        printk(KERN_ERR "general read can't execute \n");
        printk(KERN_ERR "error value = 0x%x \n", result);
        *pwr_st = 0;
    } else {
        /* only take out power state from control state */
        *pwr_st = result & CTL_STANDALONE;
        result  = 0;
    }

    return (result);
}

/******************************************************************************
 * NAME       : als_driver_read_data
 * FUNCTION   : read the value of ALS data and status in BH1780
 * REMARKS    :
 *****************************************************************************/
static int als_driver_read_data(unsigned short *data, struct i2c_client *client)
{
    int            result;
    unsigned short als_data;

    /* read register to BH1780 via i2c */
    result = i2c_smbus_read_word_data(client, REG_ALSDATA);
    if (result < 0) {
        printk(KERN_ERR "data read can't execute \n");
        printk(KERN_ERR "error value = 0x%x \n", result);
        *data = 0;
    } else {
        /* only take out power state from control state */
        als_data = (unsigned short)result;
        *data    = CONVERT_TO_BE(als_data);
        result   = 0;
    }

    return (result);
}

/******************************************************************************
 * NAME       : als_driver_general_read
 * FUNCTION   : read general multi bytes
 * REMARKS    :
 *****************************************************************************/
static int als_driver_general_read(GENREAD_ARG data, struct i2c_client *client)
{
    int            result;
    struct i2c_msg msg[2];

    if (data.size == 0) {
        return (-EINVAL);
    }
    /* check the parameter of register */
    if ((data.adr_reg < REG_ALSCONTROL) || (data.adr_reg > REG_ALSDATA_H)) {
        return (-EINVAL);
    }
    if ((data.adr_reg > REG_ALSCONTROL) && (data.adr_reg < REG_PARTID)) {
        return (-EINVAL);
    }

    /* write start address */
    msg[0].addr  = client->addr;
    msg[0].flags = 0;
    msg[0].len   = 1;
    msg[0].buf   = &data.adr_reg;

    /* read data */
    msg[1].addr  = client->addr;
    msg[1].flags = I2C_M_RD;
    msg[1].len   = data.size;
    msg[1].buf   = data.addr;

    /* block read */
    result = i2c_transfer(client->adapter, msg, 2);
    if (result < 0) {
        printk(KERN_ERR "transfer error \n");
    }

    return (result);
}

MODULE_DESCRIPTION("ROHM Ambient Light Sensor Driver");
MODULE_LICENSE("GPL");

module_init(als_init);
module_exit(als_exit);
